/*

Copyright 2001 
by Constantin Loizides <constantin.loizides@isg.de>


This file is part of the journaling-filesystem fragmentation project.
You find the files belonging to that project in the Makefile
coming together with this file.

Source code concerning that project is free software; 
you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
USA

*/

/*
	Thanks to Utz Lehmann <xfs@s2y4n2c.de>
	fibmap is now changed to handle:
		-different blocksizes
		-use of fib_res=ioctl(fd,FIGETBSZ,&fib_bsize);
		-holes
*/


/*	
	INSTALL
	check/set the correct compiler, libs and includes for your system
	in the Makefile, to compile just type ''make fibmap'' 
	Remember to have root rights when using fibmap!
   
	fibmap -h will give you information of options
*/

#include "common.h"
#include "cl_hist.h"
#include "cl_array.h"

#define PROGNAME "fibmap"
#define VERSION "0.4"

#define DEBUG_


/*for common.h */
int verbose=0;
int blocksize=BUFFERSIZE;
char *dummybuffer;

/* additional options */
int nostats=0;
long blocksgreater=0;

/* to be found out */
long f_bsize=-1; 
long fold_bsize=-1; 
long fib_bsize=-1;

/* just for curiosity */
unsigned long gcounterdirs=0;
unsigned long long gcounterfiles=0;
unsigned long gcounterslinks=0;
unsigned long gcounterelses=0;

/* for internal frag */
unsigned long long gcountbytes=0;
unsigned long long gcountfsblocks=0;
unsigned long long gcounttailsintbytes=0;

/* for external frag */
float gfrag0=0;
float gfragneg=0;
float gfrag1=0;
float gfrag10=0;
float gfrag100=0;
float gfrag1000=0;
float gfragmax=0;
float gfragpath=0;

unsigned long long gcountpathblocks=0;
unsigned long long gcountpath=0;

/* histos for frag and fragpath */
cl_Hist2<unsigned long long,float,unsigned long> fhistfrag1(12,0,1,20,2,0,2);
cl_Hist2<unsigned long long,unsigned long,unsigned long> fhistfrag2(12,10,0,20,2,0,3);

/* preallocate 1000 elements for inodes with hardlinks > 1 */
cl_Array<unsigned long> cinodes(1000);

void usage(){
	printf("Usage of %s v%s\n","fibmap","0.2");
	printf("\t -b blks\twork only on blocks greater than blks\n");
	printf("\t -h,-?\t\tthis text\n");
	printf("\t -s\t\tdon't show stats after fibmaping\n");
	printf("\t -v\t\tincrease verbosity by one\n");
	printf("\t\t\toutput is as follows:\n\t\t\tname size_in_bytes blocks_total - blocknumbers - f1 f10 f100 f1000 - fragmax - fragpath1 fragpath2\n");
	exit(1);
}

int fatal(char *mess)
{
	fprintf(stderr, "Error: %s\n", mess); 
	exit(1);
}

int rekdir_func_stat(const char* path, const struct stat *sfs)
{

	long fib_res=0,i=0;
	long fib_blk=0,fib_blk_old=0;
	long dummy=0,step=0,blockseff=0;
	long filesize,fs_blocks,fib_blocks;
	int noofhardlinks=0;
			
	unsigned long counterneg=0;
	unsigned long counter0=0;
	unsigned long counter1=0;
	unsigned long counter10=0;
	unsigned long counter100=0;
	unsigned long counter1000=0;
	unsigned long counterpath=0;
	unsigned long counterholes=0;
	unsigned long countertails=0;
	long countermin=0,countermax=0;
	
	float frag0=1.0;
	float fragneg=1.0;
	float frag1=1.0;
	float frag10=1.0;
	float frag100=1.0;
	float frag1000=1.0;
	float fragmax=1.0;
	float fragpath=1.0;

	if(S_ISDIR(sfs->st_mode)==1) {
		gcounterdirs++;	
		if(verbose>2) printf("Working in directory: %s\n",path);
	}
	else if(S_ISREG(sfs->st_mode)==1)
	{
		int fd=-100;
		
		if(f_bsize==-1){
			fd=open(path,O_RDONLY);
			if(fd==-1) return -1;

			struct statfs stf;	/*for statfs*/
			if (fstatfs(fd,&stf)<0){
				close(fd);
				return -1;
			}
			f_bsize=stf.f_bsize;
			if((fold_bsize!=-1)&&(fold_bsize!=f_bsize))
			{
				/* Warning: the blockssize dont match */
				fprintf(stderr,"Blocksizes of different directories (file systems) don't match: new %lu old %lu\n",f_bsize,fold_bsize);				
				return -1;
			}
			fib_res=ioctl(fd,FIGETBSZ,&fib_bsize);
			if(fib_res==-1)
			{
				fprintf(stderr,"FIBGETBSZ Error: %s",strerror(errno));				
				close(fd);
				return -1;
			}
#ifdef DEBUG
				printf("\nSet fssize: %lu and fibsize %lu\n",f_bsize,fib_bsize);
#endif
			close(fd);
		}

		filesize=sfs->st_size;
		fs_blocks=filesize/f_bsize+((filesize%f_bsize)?1:0);
		fib_blocks=filesize/fib_bsize+((filesize%fib_bsize)?1:0);
		step=f_bsize/fib_bsize;

#ifdef DEBUG
		printf("\nsize: %lu used fsblocks %lu and fibblocks %lu\n",filesize,fs_blocks,fib_blocks);
#endif
		if(fs_blocks<blocksgreater) return 0;

		noofhardlinks=sfs->st_nlink;
		if(noofhardlinks>1){

			unsigned long ino=sfs->st_ino;
			unsigned int i=0;
			unsigned int j=cinodes.length();
#ifdef DEBUG
			printf("Hardlink %s: %ld %d\n",path,ino,noofhardlinks);
#endif
			while(i<j)
			{
				if(cinodes[i]==ino) break;
				i++;
			}				
			if(i==j) {
				cinodes.append(&ino,1);
			}
			else {
				if(verbose>2) printf("%s, inode %lu with hardlinks %d, done before\n",path,ino,noofhardlinks);
				return 0;
			}
		}

		fd=open(path,O_RDONLY);
		if(fd==-1) return -1;

		if(verbose){
			printf("%s ", path);
			printf("%ld ",filesize);
			printf("%ld - ",fs_blocks);
		}

		for(i=0;i<fib_blocks;i+=step){
			fib_blk=i;

			fib_res=ioctl(fd,FIBMAP,&fib_blk);
			if(fib_res==-1)
			{
				close(fd);
				fprintf(stderr,"FIBMAP Error: %s",strerror(errno));				
				return -1;
			}
			fib_blk/=step;
			
			if(verbose>1){
				if((fib_blk<=0)&&(i+step<fib_blocks)) printf("h ");
				else if((fib_blk==0)&&(i+step>=fib_blocks)) printf("t ");
				else printf("%lu ",fib_blk);
			} 

			//calculate the different frags
			if((fib_blk<=0)&&(i+step<fib_blocks)){
				counterholes++;
				continue;
			}
			else if((fib_blk==0)&&(i+step>=fib_blocks)) countertails++;
				
			if(i==0){
				fib_blk_old=fib_blk;
				countermax=fib_blk;
				countermin=fib_blk;
				continue;
			}

			if (fib_blk==0){
				counter0++;
				//continue;
				dummy=1; /*count zeroes as counter0 step, means it is 1*/
			} else {
				if(countermax<fib_blk) countermax=fib_blk;
				else if(countermin>fib_blk) countermin=fib_blk;
					
				dummy=(fib_blk-fib_blk_old);
			}
				
			counterpath+=labs(dummy);
										
			if (dummy<0) counterneg++;
			else {
 				if (dummy==1) counter1++;
				if (dummy<=10) counter10++;
				if (dummy<=100) counter100++;
				if (dummy<=1000) counter1000++;
			}
			fib_blk_old=fib_blk;
		}

		close(fd);

#ifdef DEBUG
		printf("counter1: %lu counter10 %lu counter100 %lu counter1000 %lu\n",counter1,counter10,counter100,counter1000);
		printf("counter0: %lu counterpath %lu countermin %lu countermax %lu\n",counter0,counterpath,countermin,countermax);
		printf("counterneg: %lu counterholes %lu countertails %lu\n",counterneg,counterholes,countertails);
#endif

		blockseff=fs_blocks-counterholes;
		if (blockseff>1){
			countermax-=countermin;
			countermax+=1+counter0;
			fragmax=(float)countermax/(blockseff);

			frag1=(float)counter1/(blockseff-1);
			frag10=(float)counter10/(blockseff-1);
			frag100=(float)counter100/(blockseff-1);
			frag1000=(float)counter1000/(blockseff-1);

			frag0=(float)counter0/(blockseff-1);
			fragneg=(float)counterneg/(blockseff-1);
			fragpath=(float)counterpath/(blockseff-1);

			gcountpath+=counterpath;					
			gcountpathblocks+=(blockseff-1);					
		} 

		/*internal frag*/
		gcountbytes+=filesize-counterholes*f_bsize;
		gcountfsblocks+=(blockseff-countertails);
		if(countertails>0) gcounttailsintbytes+=(filesize % f_bsize);
		
		fhistfrag1.add(frag1,blockseff);
		fhistfrag2.add((long unsigned int)fragpath,blockseff);

	 	gfragneg+=fragneg;
	 	gfrag0+=frag0;
	 	gfrag1+=frag1;
	 	gfrag10+=frag10;
	 	gfrag100+=frag100;
	 	gfrag1000+=frag1000;
	 	gfragmax+=fragmax;
	 	gfragpath+=fragpath;

		gcounterfiles++;
		
#ifdef DEBUG
		if(verbose) printf("- Measure: %.3f %.3f %.3f %.3f - %.3f - %lu %lu - %.3f\n",frag1,frag10,frag100,frag1000,fragmax,counterneg,counter0,fragpath);
#else
		if(verbose) printf("- Measure: %.3f %.3f %.3f %.3f - %.3f - %.3f\n",frag1,frag10,frag100,frag1000,fragmax,fragpath);
#endif
	}
	else if(S_ISLNK(sfs->st_mode)==1){
		gcounterslinks++;
		if(verbose>2) printf("Omitting slink: %s\n",path);
	}
	else {
		gcounterelses++;
		if(verbose>2) printf("Omitting: %s\n",path);
	}
	return 0;
} 

int main(int argc, char **argv) 
{
	/*parse options*/
	char ch;
	int temp;
	while((ch=getopt(argc,argv,"b:h?sv"))!=EOF){
		switch(ch)	{	
			case 'b':
				temp=atoi(optarg);
				if(temp<1)
				{
					fprintf(stderr, "Can't change blocksgreater to \"%s\".\n", optarg);
					usage();
				} else blocksgreater=temp;
				break;

			case '?':
			case 'h': usage(); break;
			case 's': nostats=1; break;
			case 'v': verbose++; break;
			default: usage(); break;
		}
	}

	for(;optind<argc;optind++)
	{                       
#ifdef DEBUG
		printf("\nWorking on: \"%s\"\n", argv[optind]);
#endif

		fold_bsize=f_bsize;
		f_bsize=-1; 
		fib_bsize=-1;

		if (walk_through_dirs(argv[optind],rekdir_func_stat)==-1) fatal("ret_walk_through_dirs failed...");
	}

	
#ifdef DEBUG
	printf("Dirs %lu\nFiles %lu\nSlinks %lu\nElses %lu\n",gcounterdirs,gcounterfiles,gcounterslinks,gcounterelses);
#endif

	/* Statistics */
	if(gcounterfiles<=1) exit(0);
	if(nostats==1) exit(0);
	
 	gfrag0/=gcounterfiles;
 	gfrag1/=gcounterfiles;
 	gfragneg/=gcounterfiles;;
 	gfrag10/=gcounterfiles;;
 	gfrag100/=gcounterfiles;;
 	gfrag1000/=gcounterfiles;;
 	gfragmax/=gcounterfiles;;
 	gfragpath/=gcounterfiles;;
	float intfrag=(float)(gcountbytes)/(f_bsize*gcountfsblocks+gcounttailsintbytes);
	float gfragpath2=1.;
	if(gcountpathblocks>0) gfragpath2=(float)gcountpath/gcountpathblocks;

#ifdef DEBUG
	printf("Result for %lu files %lu dirs: Int: %.3f Ext: %.3f %.3f %.3f %.3f - %.3f - Path %.3f %.3f\n",gcounterfiles,gcounterdirs,intfrag,gfrag1,gfrag10,gfrag100,gfrag1000,gfragmax,gfragpath,gfragpath2);
#else
	printf("Result for %llu files %lu dirs: Int: %.3f Ext: %.3f %.3f %.3f %.3f - %.3f - Path %.3f %.3f\n",gcounterfiles,gcounterdirs,intfrag,gfrag1,gfrag10,gfrag100,gfrag1000,gfragmax,gfragpath,gfragpath2);
	printf("Mean Internal frag: %.6f\nMean External frag: %.6f\nMean Frag. Path %.3f\n",intfrag,gfrag1,gfragpath);
#endif
	printf("Hist for frag\n");
	cout << fhistfrag1 << endl;
	printf("Hist for frag path\n");
	cout << fhistfrag2 << endl;
}

